from . import acme
from . import acmesession
from .servers import Tokens, start_servers
from .utils import fatal, yesno as _yesno
from .utils import _file_generator
from .utils import _dated_file_generator
from functools import partial
from pathlib import Path
import OpenSSL
import click
import datetime
import json
import subprocess
import time


OPENSSL = 'openssl'
LETSENCRYPT_CERT = 'lets-encrypt-x3-cross-signed'


def genkey(fn, yesno, ask=False, keylen=4096):
    if ask:
        click.echo('There is no user key in the current directory %s.' % fn.parent)
        if not yesno('Do you want to create a user key?', default=False):
            fatal('No user key created')
    subprocess.check_call([
        OPENSSL, 'genrsa', '-out', str(fn), str(keylen)])


def genpub(fn, key):
    subprocess.check_call([
        OPENSSL, 'rsa', '-in', str(key), '-pubout', '-out', str(fn)])


def createSubjectAltName(domains):
    return 'subjectAltName = %s' % ','.join('DNS:%s' % x for x in domains)


def gencsr(fn, key, domains):
    if len(domains) > 1:
        config_fn = fn.parent.joinpath('openssl.cnf')
        with config_fn.open('wb') as config:
            lines = [
                '[ req ]',
                'distinguished_name  = req_distinguished_name',
                '',
                '[ req_distinguished_name ]',
                '',
                '[SAN]',
                createSubjectAltName(domains),
                '']
            config.write(bytes('\n'.join(lines).encode('ascii')))
        subprocess.check_call([
            OPENSSL, 'req', '-sha256', '-new',
            '-key', str(key), '-out', str(fn), '-subj', '/',
            '-reqexts', 'SAN', '-config', str(config_fn)])
    else:
        subprocess.check_call([
            OPENSSL, 'req', '-sha256', '-new',
            '-key', str(key), '-out', str(fn),
            '-subj', '/CN=%s' % domains[0]])


def verify_domains(cert_or_req, domains):
    names = set()
    subject = dict(cert_or_req.get_subject().get_components()).get(
        b'CN', b'').decode('ascii')
    if subject:
        names.add(subject)
    if hasattr(cert_or_req, 'get_extensions'):
        extensions = cert_or_req.get_extensions()
    else:
        extensions = [
            cert_or_req.get_extension(x)
            for x in range(cert_or_req.get_extension_count())]
    for ext in extensions:
        if ext.get_short_name() != b'subjectAltName':
            continue
        alt_names = [
            x.strip().replace('DNS:', '')
            for x in str(ext).split(',')]
        names = names.union(alt_names)
    unmatched = set(domains).symmetric_difference(names)
    if unmatched:
        click.echo(click.style(
            "Unmatched alternate names %s" % ', '.join(unmatched), fg="red"))
        return False
    return True


def verify_csr(csr, domains):
    subprocess.check_call([
        OPENSSL, 'req', '-noout', '-verify', '-in', str(csr)])
    with csr.open('rb') as f:
        req = OpenSSL.crypto.load_certificate_request(
            OpenSSL.crypto.FILETYPE_PEM, f.read())
    if not verify_domains(req, domains):
        subprocess.check_call([
            OPENSSL, 'req', '-noout', '-text', '-in', str(csr)])
        return False
    return True


def gender(fn, csr):
    subprocess.check_call([
        OPENSSL, 'req', '-outform', 'DER', '-out', str(fn), '-in', str(csr)])


def check_acme_registration(genreg, jwk, file_generator):
    click.echo("Checking registration at letsencrypt.")
    reg_fn = file_generator(
        'registration info', '.json', genreg)
    with reg_fn.open() as f:
        registration_info = json.load(f)
    click.echo("Registered on %s via %s" % (
        registration_info["createdAt"], registration_info["initialIp"]))
    click.echo("Contact: %s" % ", ".join(registration_info["contact"]))
    if 'agreement' in registration_info:
        click.echo("Agreement: %s" % registration_info["agreement"])
    if registration_info['key'] != jwk:
        fatal("The public user key and the registration info don't match.")


def _genreg(fn, acme, email):
    data = acme.new_account(email)
    data = json.dumps(data, sort_keys=True, indent=4)
    with fn.open('w') as out:
        out.write(data)


def gencrt(fn, acme_factory, check_registration, der, user_pub, email, domains):
    with der.open('rb') as f:
        der_data = f.read()
    acme = acme_factory()
    genreg = partial(_genreg, acme=acme, email=email)
    check_registration(genreg)
    click.echo("Preparing challenges for %s." % ', '.join(domains))
    (order_uri, info) = acme.handle_order(domains)
    if info['status'] == 'pending':
        for authorization in info['authorizations']:
            acme.handle_authorization(authorization)
        count = 0
    while info['status'] == 'pending':
        while count < 5:
            authorizations_ok = [
                acme.tokens.get_status(authorization) in ('requested', 'valid')
                for authorization in info['authorizations']]
            if all(authorizations_ok):
                break
            click.echo(".", nl=False)
            time.sleep(1)
            count += 1
        info = acme.poll_order(order_uri)
    if info.get('status') == 'ready':
        info = acme.finalize_order(info['finalize'], der_data)
    if info.get('status') == 'valid':
        cert = acme.get_certificate(info['certificate'])
        with fn.open('wb') as out:
            out.write(cert)
    else:
        fatal("Failed to get certificate.")


def verify_crt(crt, domains):
    with crt.open('rb') as f:
        cert = OpenSSL.crypto.load_certificate(
            OpenSSL.crypto.FILETYPE_PEM, f.read())
    issuer = dict(cert.get_issuer().get_components()).get(
        b'CN', b'unknown').decode('ascii')
    if issuer in ["happy hacker fake CA",
                  "Fake LE Intermediate X1"]:
        click.echo(click.style("Certificate issued by staging CA!", fg="red"))
    elif issuer in ["Let's Encrypt Authority X1",
                    "Let's Encrypt Authority X2",
                    "Let's Encrypt Authority X3",
                    "Let's Encrypt Authority X4"]:
        click.echo(click.style("Certificate issued by: %s" % issuer, fg="green"))
    else:
        click.echo(click.style("Unknown CA: %s" % issuer, fg="red"))
    if not verify_domains(cert, domains):
        subprocess.check_call([
            OPENSSL, 'x509', '-noout', '-text', '-in', str(crt)])
        return False
    return True


def chain(fn, crt, pem):
    with fn.open('wb') as out:
        for name in (crt, pem):
            with name.open('rb') as f:
                data = f.read()
                out.write(data)
                if not data.endswith(b'\n'):
                    out.write(b'\n')


def remove(yesno, base, *patterns):
    files = []
    for pattern in patterns:
        for fn in base.glob(pattern):
            files.append(fn)
            click.echo(fn.relative_to(base))
    if not yesno("Do you want to remove the above invalid files for a clean retry?"):
        fatal('Aborted.')
    for fn in files:
        if fn.exists():
            fn.unlink()


def generate(base, main, acme_factory, acme_uris_factory, domains, file_gens, yesno):
    user_key = file_gens['file'](base, 'user')(
        'private user key', '.key', genkey, yesno=yesno, ask=True)
    user_pub = file_gens['file'](base, 'user')(
        'public user key', '.pub', genpub, user_key)
    assert user_pub.parent == base
    key_base = base.joinpath(main)
    if not key_base.exists():
        key_base.mkdir()
    key = file_gens['dated'](key_base, main)(
        'key', '.key', genkey, yesno=yesno)
    date_gen = file_gens['current'](key_base, main)
    while True:
        csr = date_gen('csr', '.csr', gencsr, key, domains)
        if verify_csr(csr, domains):
            break
        remove(yesno, key_base, '*.csr', '*.crt', '*.der')
    with user_key.open('rb') as f:
        priv = OpenSSL.crypto.load_privatekey(
            OpenSSL.crypto.FILETYPE_PEM, f.read())
    jwk = acmesession.get_jwk(user_pub)
    session = acmesession.get_session(jwk, priv)
    check_registration = partial(
        check_acme_registration,
        jwk=jwk,
        file_generator=file_gens['registration'](base, 'registration'))
    while True:
        der = date_gen('der', '.der', gender, csr)
        assert der.parent == key_base
        acme_factory = partial(
            acme_factory,
            acme_uris=acme_uris_factory(session=session))
        crt = date_gen('chained crt', '-chained.crt', gencrt, acme_factory, check_registration, der, user_pub, base.name, domains)
        if verify_crt(crt, domains):
            break
        remove(yesno, key_base, '*.crt', '*.der')


def domain_key(x):
    return (len(x), x)


@click.command(context_settings=dict(help_option_names=['-h', '--help']))
@click.option(
    "--dns/--no-dns", default=False,
    help="Enable/disable DNS challenge. HTTP challenge is tried first if enabled.")
@click.option(
    "--http/--no-http", default=True,
    help="Enable/disable HTTP challenge.")
@click.option(
    "-r/-R", "--regenerate/--dont-regenerate", default=False,
    help="Force creating a new certificate even if one for the current day exists.")
@click.option(
    "-s/-p", "--staging/--production", default=False,
    help="Use staging server of letsencrypt.org for testing.")
@click.option(
    "-u", "--update", metavar="PATH",
    help="Update the certificates in PATH.")
@click.option(
    "--update-registration/--no-update-registration", default=False,
    help="Force an update of the registration, for example to agree to newer terms of service.")
@click.option(
    "-y", "--yes/--ask", default=False,
    help="Use staging server of letsencrypt.org for testing.")
@click.argument("domains", metavar="[DOMAIN]...", nargs=-1)
def main(domains, dns, http, regenerate, staging, update, update_registration, yes):
    """Creates a certificate for one or more domains.

    By default a new certificate is generated, except when running again on
    the same day."""
    yesno = partial(_yesno, always_yes=yes)
    if staging:
        ca = "https://acme-staging-v02.api.letsencrypt.org"
    else:
        ca = "https://acme-v02.api.letsencrypt.org"
    base = Path.cwd()
    date = datetime.date.today().strftime("%Y%m%d")
    current = 'force' if regenerate else True
    file_gens = dict(
        file=_file_generator,
        dated=partial(_dated_file_generator, date=date),
        current=partial(_dated_file_generator, date=date, current=current),
        registration=partial(_file_generator, update=update_registration))
    cli_domains = sorted(domains, key=domain_key)
    challenges = ['http-01']
    if http and 'http-01' not in challenges:
        challenges.append('http-01')
    if not http and 'http-01' in challenges:
        challenges.remove('http-01')
    if dns and 'dns-01' not in challenges:
        challenges.append('dns-01')
    if not dns and 'dns-01' in challenges:
        challenges.remove('dns-01')
    if not challenges:
        fatal("No challenge types enabled.")
    options = dict(challenges=challenges)
    if update:
        path = Path(update).absolute()
        if not path.exists():
            fatal("The path %s doesn't exist." % update)
        base = path.parent
        options_path = path.joinpath('options.json')
        if options_path.exists():
            with options_path.open() as f:
                options.update(json.load(f))
        else:
            options['domains'] = [path.name] + [
                x.name.rsplit('.authz_info.json', 1)[0]
                for x in path.glob('*.authz_info.json')]
            options['main'] = path.name
    option_domains = tuple(
        sorted(set(options.get('domains', [])), key=domain_key))
    if option_domains:
        click.echo(click.style(
            "Existing domains from '%s': %s" % (
                update, ", ".join(option_domains)),
            fg="green"))
        domains = tuple(set(domains).union(option_domains))
    if cli_domains:
        click.echo(click.style(
            "Domains from command line: %s" % ", ".join(cli_domains),
            fg="green"))
        domains = tuple(set(domains).union(cli_domains))
    domains = sorted(domains, key=domain_key)
    main = options.get('main', domains[0] if domains else None)
    tokens = Tokens()
    start_servers(challenges, tokens)
    if domains:
        click.echo(click.style(
            "Domains to update: %s" % ", ".join(domains),
            fg="green"))
        click.echo(click.style(
            "Main domain: %s" % main,
            fg="green"))
        click.echo(click.style(
            "Challenges: %s" % ", ".join(challenges),
            fg="green"))
        if update:
            if not yesno("Do you want to update with the above settings?"):
                fatal('Aborted.')
        acme_factory = partial(
            acme.ACME,
            challenges=challenges,
            tokens=tokens,
            yesno=yesno)
        acme_uris_factory = partial(acme.ACMEUris, ca=ca)
        generate(
            base, main,
            acme_factory, acme_uris_factory,
            domains, file_gens, yesno)
        with base.joinpath(main, 'options.json').open("w") as f:
            f.write(json.dumps(
                dict(
                    challenges=challenges,
                    main=main,
                    domains=domains),
                sort_keys=True, indent=4))
    else:
        fatal(
            "No domains given.\n"
            "Use --help to print usage.")


if __name__ == '__main__':
    main()



from .acmesession import b64
from .utils import fatal, fatal_response
from functools import partial
import click


class ACMEUris:
    def __init__(self, ca, session):
        self.session = session
        res = session.get(ca + '/directory')
        content_type = res.headers.get('Content-Type')
        if res.status_code != 200 or content_type != 'application/json':
            fatal_response(
                "Couldn't get directory from CA server '%s'" % ca, res)
        self.uris = res.json()
        self.terms_uri = self.uris.get('meta', {}).get('termsOfService')
        self.new_account = partial(session.post_jwk_signed, self.uris['newAccount'])
        self.new_nonce = partial(session.head, self.uris['newNonce'])
        self.new_order = partial(session.post_kid_signed, self.uris['newOrder'])


class ACME:
    def __init__(self, acme_uris, challenges, tokens, yesno):
        self.challenges = challenges
        self.acme_uris = acme_uris
        self.tokens = tokens
        self.yesno = yesno

    def authorized_token(self, token):
        return "{}.{}".format(token, self.acme_uris.session.thumbprint).encode('ascii')

    def ensure_account(self):
        self.ensure_nonce()
        if self.acme_uris.session.kid is None:
            res = self.acme_uris.new_account(onlyReturnExisting=True)
            if res.status_code != 200:
                fatal_response("Bad account check response", res)
            self.acme_uris.session.kid = res.headers['Location']

    def ensure_nonce(self):
        if self.acme_uris.session.nonce is None:
            res = self.acme_uris.new_nonce()
            if res.status_code != 204:
                fatal_response("Bad newNonce response", res)

    def finalize_order(self, uri, der_data):
        res = self.acme_uris.session.post_kid_signed(uri, csr=b64(der_data))
        if res.status_code != 200:
            fatal_response("Bad finalize order request", res)
        return res.json()

    def get_certificate(self, uri):
        res = self.acme_uris.session.post_as_get(uri)
        if res.status_code != 200:
            fatal_response("Bad get certificate request", res)
        return res.content

    def handle_authorization(self, uri):
        res = self.acme_uris.session.post_as_get(uri)
        if res.status_code != 200:
            fatal_response("Bad authorization response", res)
        data = res.json()
        if data.get('status') == 'valid':
            self.tokens.set_status(uri, 'valid')
            return
        domain = data['identifier']['value']
        click.echo("Authorizing %s." % domain)
        for challenge_type in self.challenges:
            for challenge in data['challenges']:
                if challenge['type'] != challenge_type:
                    continue
                authorized_token = self.authorized_token(challenge['token'])
                if challenge_type == 'dns-01':
                    self.tokens.add_dns_reply(
                        uri, domain, authorized_token)
                elif challenge_type == 'http-01':
                    self.tokens.add_http_reply(
                        uri, challenge['token'], authorized_token)
                res = self.acme_uris.session.post_kid_signed(challenge['url'])
                if res.status_code != 200:
                    fatal_response("Bad challenge request %s" % challenge, res)

    def handle_order(self, domains):
        self.ensure_account()
        res = self.acme_uris.new_order(identifiers=[
            dict(type="dns", value=domain)
            for domain in domains])
        if res.status_code != 201:
            fatal_response("Bad newOrder response", res)
        return res.headers['Location'], res.json()

    def new_account(self, email):
        self.ensure_nonce()
        question = "Do you agree to the terms of service"
        if self.acme_uris.terms_uri:
            question = "%s at %s" % (question, self.acme_uris.terms_uri)
        if not self.yesno(question):
            fatal("Didn't agree to terms of service.")
        res = self.acme_uris.new_account(
            contact=["mailto:" + email],
            termsOfServiceAgreed=True)
        if res.status_code == 201:
            return res.json()
        fatal_response("Bad newAccount response", res)

    def poll_order(self, uri):
        res = self.acme_uris.session.post_as_get(uri)
        if res.status_code != 200:
            fatal_response("Bad order poll response", res)
        return res.json()



from .acmesession import b64
from .utils import fatal
import click
import dns.name
import dns.message
import dns.rdtypes.ANY.TXT
import dns.rrset
import hashlib
import http.server
import socket
import threading
import time


class Tokens(dict):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._status = {}

    def add_dns_reply(self, uri, domain, authorized_token):
        digest = hashlib.sha256(authorized_token).digest()
        txt = b64(digest)
        click.echo('_acme-challenge.%s. IN TXT "%s"' % (domain, txt))
        self[dns.name.from_text(domain)] = (uri, txt)

    def add_http_reply(self, uri, token, authorized_token):
        self[token] = (uri, authorized_token)

    def set_status(self, uri, status):
        self._status[uri] = status

    def get_status(self, uri):
        return self._status.get(uri, 'unknown')


class HTTPRequestHandler(http.server.BaseHTTPRequestHandler):
    def do_GET(self):
        parts = self.path.split('/')
        if len(parts) != 4 or parts[1:3] != ['.well-known', 'acme-challenge']:
            self.send_response(404)
            self.end_headers()
            return
        token = parts[3]
        if token not in self.server.tokens:
            self.send_response(404)
            self.end_headers()
            return
        self.send_response(200)
        self.end_headers()
        (uri, reply) = self.server.tokens[token]
        self.wfile.write(reply)
        self.server.tokens.set_status(uri, 'requested')


class DNSServer:
    def __init__(self, address):
        self.address = address

    def handle_request(self, addr, data):
        query = dns.message.from_wire(data)
        response = dns.message.make_response(query)
        for question in query.question:
            if question.rdtype != 16:
                continue
            if question.name.labels[0].lower() != b'_acme-challenge':
                continue
            domain = question.name.parent()
            print('Got request from %s for domain %s.' % (addr, domain))
            if domain not in self.tokens:
                continue
            (uri, txt) = self.tokens[domain]
            print('Answering with "%s" to %s for domain %s.' % (txt, addr, domain))
            response.answer.append(dns.rrset.from_rdata(
                question.name,
                0,
                dns.rdtypes.ANY.TXT.TXT(
                    question.rdclass, question.rdtype, [txt.encode('ascii')])))
            self.tokens.set_status(uri, 'requested')
        return response.to_wire()

    def __call__(self):
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        try:
            s.bind(self.address)
        except OSError as e:
            fatal(str(e))
        s.listen(1)
        while 1:
            conn, addr = s.accept()
            with conn:
                data = conn.recv(1024)
                reply = self.handle_request(addr, data)
                if reply is not None:
                    conn.send(reply)


def start_servers(challenges, tokens):
    if any(x.startswith('http') for x in challenges):
        address = ('localhost', 8080)
        server = http.server.HTTPServer(address, HTTPRequestHandler)
        thread = threading.Thread(target=server.serve_forever, daemon=True)
        click.echo("Starting http server on %s:%s" % address)
        thread.start()
        time.sleep(0.1)
        if not thread.is_alive():
            fatal("Failed to start HTTP server on port 8080.")
        server.tokens = tokens
    if any(x.startswith('dns') for x in challenges):
        dnsaddress = ('localhost', 8053)
        dnsserver = DNSServer(dnsaddress)
        dnsthread = threading.Thread(target=dnsserver, daemon=True)
        click.echo("Starting dns server on %s:%s" % dnsaddress)
        dnsthread.start()
        time.sleep(0.1)
        if not dnsthread.is_alive():
            fatal("Failed to start DNS server on port 8053.")
        dnsserver.tokens = tokens



from base64 import urlsafe_b64encode
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import serialization
from functools import partial
import OpenSSL
import binascii
import hashlib
import json
import requests


def b64(data):
    return urlsafe_b64encode(data).replace(b"=", b"").decode('ascii')


def _leading_zeros(arg):
    if len(arg) % 2:
        return '0' + arg
    return arg


def _encode(data):
    return b64(binascii.unhexlify(
        _leading_zeros(hex(data)[2:].rstrip('L'))))


def get_jwk(user_pub):
    backend = default_backend()
    with user_pub.open('rb') as f:
        pub = serialization.load_pem_public_key(f.read(), backend)
        pub_numbers = pub.public_numbers()
        return dict(
            kty="RSA",
            e=_encode(pub_numbers.e),
            n=_encode(pub_numbers.n))


def get_thumbprint(jwk):
    return b64(hashlib.sha256(json.dumps(
        jwk,
        sort_keys=True,
        separators=(',', ':')).encode('ascii')).digest())


def protected(header, uri, nonce, kid):
    data = dict(header)
    data['nonce'] = nonce
    data['url'] = uri
    if kid is not None:
        data.pop('jwk')
        data['kid'] = kid
    return b64(json.dumps(data, sort_keys=True, indent=4).encode('utf-8'))


def sign_sha256(sig_data, priv):
    return OpenSSL.crypto.sign(
        priv, sig_data.encode('ascii'), 'sha256')


def dumps(**kw):
    return b64(json.dumps(dict(**kw), sort_keys=True).encode('utf-8'))


def _dumps_signed(nonce, uri, header, payload, sign, kid=None):
    data = protected(header, uri, nonce, kid)
    sig_data = "%s.%s" % (data, payload)
    signature = b64(sign(sig_data))
    data = dict(
        protected=data,
        payload=payload,
        signature=signature)
    return data


class ACMESession:
    def __init__(self, dumps_signed, thumbprint):
        self.dumps_signed = dumps_signed
        self.kid = None
        self.nonce = None
        self.thumbprint = thumbprint
        self.session = requests.Session()
        self.session.hooks = dict(response=self.response_hook)
        self.get = self.session.get
        self.head = self.session.head
        self.post = self.session.post

    def response_hook(self, response, *args, **kwargs):
        if 'Replay-Nonce' in response.headers:
            self.nonce = response.headers['Replay-Nonce']

    def post_jwk_signed(self, uri, **kw):
        payload = dumps(**kw)
        return self.session.post(
            uri,
            headers={'Content-Type': 'application/jose+json'},
            json=self.dumps_signed(self.nonce, uri, payload=payload))

    def post_kid_signed(self, uri, **kw):
        payload = dumps(**kw)
        return self.session.post(
            uri,
            headers={'Content-Type': 'application/jose+json'},
            json=self.dumps_signed(self.nonce, uri, payload=payload, kid=self.kid))

    def post_as_get(self, uri):
        return self.session.post(
            uri,
            headers={'Content-Type': 'application/jose+json'},
            json=self.dumps_signed(self.nonce, uri, payload="", kid=self.kid))


def get_session(jwk, priv):
    dumps_signed = partial(
        _dumps_signed,
        header=dict(alg="RS256", jwk=dict(jwk)),
        sign=partial(sign_sha256, priv=priv))
    thumbprint = get_thumbprint(jwk)
    return ACMESession(dumps_signed, thumbprint)



from functools import partial
import pytest


@pytest.fixture
def acme_factory(ca):
    from certsling.acme import ACME
    return partial(
        ACME,
        challenges=[],
        tokens={},
        yesno=lambda m: True)


@pytest.fixture
def acme_uris_factory(ca):
    from certsling.acme import ACMEUris
    return partial(
        ACMEUris,
        ca=ca)


@pytest.fixture
def generate(acme_factory, acme_uris_factory, base):
    from certsling import generate
    return partial(
        generate,
        acme_factory=acme_factory,
        acme_uris_factory=acme_uris_factory,
        base=base)


def get_file_gens(date=None, regenerate=False, update_registration=False):
    from certsling import _file_generator, _dated_file_generator
    import datetime

    if date is None:
        date = datetime.date.today().strftime("%Y%m%d")
    current = 'force' if regenerate else True
    return dict(
        file=_file_generator,
        dated=partial(_dated_file_generator, date=date),
        current=partial(_dated_file_generator, date=date, current=current),
        registration=partial(_file_generator, update=update_registration))


def test_user_gen(base, generate, verify_crt_true):
    assert list(base.iterdir()) == []
    domains = ['example.com']
    generate(
        main=domains[0], domains=domains,
        file_gens=get_file_gens(),
        yesno=lambda m, default=None: True)
    fns = list(x.name for x in base.iterdir())
    assert 'user.key' in fns
    assert 'user.pub' in fns



import base64
import binascii
import http.server
import json
import pytest
import threading


def d64(data):
    for pad in ('', '=', '=='):
        try:
            return base64.urlsafe_b64decode(data + pad)
        except binascii.Error:
            pass


class HTTPRequestHandler(http.server.BaseHTTPRequestHandler):
    def make_url(self, path):
        addr = "http://%s:%s/{0}" % self.server.socket.getsockname()
        return addr.format(path)

    def do_GET(self):
        if self.path == '/directory':
            self.send_response(200)
            self.send_header('Replay-Nonce', 'nonce')
            methods = ['newAccount', 'newNonce', 'newOrder']
            self.write_response({x: self.make_url(x) for x in methods})
        else:
            raise ValueError("GET %s" % self.path)

    def do_HEAD(self):
        if self.path == '/newNonce':
            self.send_response(200)
            self.send_header('Replay-Nonce', 'nonce')
            self.end_headers()
        else:
            raise ValueError("POST %s" % self.path)

    @property
    def data(self):
        if not hasattr(self, '_data'):
            data = self.rfile.read(int(self.headers['Content-Length']))
            self._data = json.loads(data.decode('utf-8'))
        return self._data

    def get_payload(self):
        payload = d64(self.data['payload'])
        return json.loads(payload.decode('utf-8'))

    def get_protected(self):
        protected = d64(self.data['protected'])
        return json.loads(protected.decode('utf-8'))

    def write_response(self, response):
        response = json.dumps(response)
        self.send_header('Content-Type', 'application/json')
        self.end_headers()
        self.wfile.write(response.encode('ascii'))

    def do_POST(self):
        if self.path == '/newOrder':
            payload = self.get_payload()
            self.send_response(201)
            self.send_header('Location', '')
            self.write_response({
                'status': 'valid',
                'certificate': self.make_url('certificate')})
        elif self.path == '/newAccount':
            payload = self.get_payload()
            protected = self.get_protected()
            if 'onlyReturnExisting' in payload:
                if protected['jwk']['n'] in self.server.accounts:
                    self.send_response(200)
                    self.send_header('Location', self.make_url('account/1'))
                    self.end_headers()
                else:
                    self.send_response(400)
                    self.end_headers()
            else:
                self.send_response(201)
                self.write_response({
                    'createdAt': 'createdAt',
                    'initialIp': 'initialIp',
                    'contact': payload['contact'],
                    'key': protected['jwk']})
                self.server.accounts[protected['jwk']['n']] = True
        elif self.path == '/certificate':
            self.send_response(200)
            self.end_headers()
        else:
            raise ValueError("POST %s" % self.path)

    def log_request(self, code):
        return


@pytest.fixture
def server():
    address = ('localhost', 0)
    server = http.server.HTTPServer(address, HTTPRequestHandler)
    server.accounts = {}
    thread = threading.Thread(target=server.serve_forever, daemon=True)
    thread.start()
    yield server
    server.shutdown()


@pytest.fixture
def ca(server):
    return "http://%s:%s" % server.socket.getsockname()


@pytest.fixture(autouse=True)
def genkey(monkeypatch):
    from certsling import genkey
    from functools import partial
    genkey_partial = partial(genkey, keylen=512)
    monkeypatch.setattr("certsling.genkey", genkey_partial)
    return genkey_partial


@pytest.fixture
def base(tmpdir):
    from pathlib import Path
    return Path(tmpdir.ensure_dir('foo@example.com').strpath)


@pytest.fixture
def verify_crt_true(monkeypatch):
    monkeypatch.setattr("certsling.verify_crt", lambda *x: True)



from functools import partial
import base64
import pytest
import subprocess


@pytest.fixture()
def server_key(base, genkey):
    fn = base.joinpath('server.key')
    genkey(fn, yesno=lambda m: True)
    return fn


@pytest.fixture
def gencsr(base, server_key):
    from certsling import gencsr
    return partial(gencsr, key=server_key)


@pytest.fixture()
def ca_key(base, genkey):
    fn = base.joinpath('ca.key')
    genkey(fn, yesno=lambda m: True)
    return fn


@pytest.fixture(params=[
    'Foo',
    'happy hacker fake CA',
    'Fake LE Intermediate X1',
    "Let's Encrypt Authority X1",
    "Let's Encrypt Authority X2",
    "Let's Encrypt Authority X3",
    "Let's Encrypt Authority X4"])
def ca_crt(base, ca_key, request):
    fn = base.joinpath('ca.crt')
    subprocess.check_call([
        'openssl', 'req',
        '-new', '-x509', '-key', str(ca_key), '-out', str(fn),
        '-subj', '/C=DE/CN=%s' % request.param,
        '-days', '90'])
    return fn


@pytest.fixture()
def signer(base, ca_key, ca_crt):
    from certsling import createSubjectAltName

    def signer(csr, crt, domains):
        import textwrap
        conf = base.joinpath('ca.conf')
        conf.open('w').write(textwrap.dedent("""\
        extensions = extend
        [extend] # openssl extensions
        %s
        """ % createSubjectAltName(domains)))
        subprocess.check_call([
            'openssl', 'x509', '-extfile', str(conf),
            '-req', '-CA', str(ca_crt), '-CAkey', str(ca_key),
            '-set_serial', '01', '-in', str(csr), '-out', str(crt),
            '-days', '90'])
    return signer


@pytest.mark.parametrize(['csr_domains', 'verify_domains', 'csr_result', 'crt_result'], [
    (
        ['example.com'],
        ['example.com'],
        True, True),
    (
        ['example.com', 'foo.example.com'],
        ['example.com', 'foo.example.com'],
        True, True),
    (
        ['example.com', 'other.org'],
        ['example.com', 'other.org'],
        True, True),
    (
        ['example.com', 'foo.example.com'],
        ['example.com'],
        False, False),
    (
        ['example.com'],
        ['example.com', 'foo.example.com'],
        False, False)])
def test_verify(base, crt_result, csr_domains, csr_result, gencsr, signer, verify_domains):
    from certsling import verify_crt, verify_csr
    csr_fn = base.joinpath('domain.csr')
    gencsr(csr_fn, domains=csr_domains)
    csr_content = csr_fn.open('r').read().splitlines()
    csr_content = base64.b64decode('\n'.join(csr_content[1:-1]))
    for domain in csr_domains:
        assert domain.encode('ascii') in csr_content
    assert verify_csr(csr_fn, domains=verify_domains) is csr_result
    crt_fn = base.joinpath('domain.crt')
    signer(str(csr_fn), str(crt_fn), csr_domains)
    crt_content = crt_fn.open('r').read().splitlines()
    crt_content = base64.b64decode('\n'.join(crt_content[1:-1]))
    for domain in csr_domains:
        assert domain.encode('ascii') in crt_content
    assert verify_crt(crt_fn, domains=verify_domains) is crt_result



import click
import json
import sys


def fatal(msg, code=3):
    click.echo(click.style(msg, fg='red'))
    sys.exit(code)


def fatal_response(msg, response, code=3):
    try:
        data = response.json()
    except ValueError:
        data = {}
    headers = '\n'.join(': '.join(x) for x in response.headers.items())
    fatal("%s: %s %s\n%s\n%s" % (
        msg,
        response.status_code, response.reason,
        json.dumps(data, sort_keys=True, indent=4),
        headers))


def ensure_not_empty(fn):
    if fn.exists():
        with fn.open('rb') as f:
            l = len(f.read().strip())
        if l:
            return True
        fn.unlink()
    return False


def _file_generator(base, name, update=False):
    def generator(description, ext, generate, *args, **kw):
        fn = base.joinpath("%s%s" % (name, ext))
        rel = fn.relative_to(base)
        if not update and ensure_not_empty(fn):
            click.echo(click.style(
                "Using existing %s '%s'." % (description, rel), fg='green'))
            return fn
        click.echo("Writing %s '%s'." % (description, rel))
        generate(fn, *args, **kw)
        return fn
    return generator


def _dated_file_generator(base, name, date, current=False):
    def generator(description, ext, generate, *args, **kw):
        fn = base.joinpath("%s%s" % (name, ext))
        rel = fn.relative_to(base)
        fn_date = base.joinpath("%s-%s%s" % (name, date, ext))
        rel_date = fn_date.relative_to(base)
        if current == 'force' and fn.exists():
            click.echo("Unlinking existing %s '%s'." % (description, rel))
            fn.unlink()
        if ensure_not_empty(fn):
            if not current or fn.resolve() == fn_date:
                click.echo(click.style(
                    "Using existing %s '%s'." % (description, rel), fg='green'))
                return fn
            elif fn.exists():
                click.echo("Unlinking existing %s '%s'." % (description, rel))
                fn.unlink()
        if not ensure_not_empty(fn):
            click.echo("Generating %s '%s'." % (description, rel_date))
            generate(fn_date, *args, **kw)
        if fn_date.exists():
            click.echo("Linking %s '%s'." % (description, rel_date))
            fn.symlink_to(fn_date.name)
        return fn
    return generator


def yesno(question, default=None, all=False, always_yes=False):
    if default is True:
        question = "%s [Yes/no" % question
        answers = {
            False: ('n', 'no'),
            True: ('', 'y', 'yes'),
        }
    elif default is False:
        question = "%s [yes/No" % question
        answers = {
            False: ('', 'n', 'no'),
            True: ('y', 'yes'),
        }
    else:
        question = "%s [yes/no" % question
        answers = {
            False: ('n', 'no'),
            True: ('y', 'yes'),
        }
    if all:
        if default == 'all':
            answers['all'] = ('', 'a', 'all')
            question = "%s/All" % question
        else:
            answers['all'] = ('a', 'all')
            question = "%s/all" % question
    question = "%s] " % question
    while 1:
        if default is None and always_yes:
            answer = 'yes'
        else:
            answer = input(question).lower()
        for option in answers:
            if answer in answers[option]:
                return option
        if all:
            click.echo("You have to answer with y, yes, n, no, a or all.")
        else:
            click.echo("You have to answer with y, yes, n or no.")



from setuptools import setup
import os


README = open(os.path.abspath('README.rst')).read()
HISTORY = open(os.path.abspath('HISTORY.rst')).read()


setup(
    name='certsling',
    version='0.9.0',
    description='Opinionated letsencrypt acme client working via a ssh port forward.',
    long_description="\n\n".join([README, HISTORY]),
    url='https://github.com/fschulze/certsling',
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: System Administrators",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8"],
    install_requires=[
        'click',
        'dnspython>=1.15.0',
        'pyOpenSSL',
        'requests'],
    entry_points={
        'console_scripts': ['certsling = certsling:main']},
    packages=['certsling'],
    package_dir={'': 'src'},
    python_requires='>=3.5')



def test_dns():
    from certsling.servers import DNSServer
    from certsling.servers import Tokens
    import dns.name
    request = b'\xed:\x00\x10\x00\x01\x00\x00\x00\x00\x00\x01\x0f_acMe-cHaLlenge\tLOcaLHOSt\x07ExaMPle\x03com\x00\x00\x10\x00\x01\x00\x00)\x10\x00\x00\x00\x80\x00\x00\x00'
    server = DNSServer(('127.0.0.1', 8053))
    server.tokens = Tokens()
    server.tokens[dns.name.from_text('localhost.example.com')] = ('uri', 'foo')
    reply = server.handle_request('192.168.1.42', request)
    assert reply == b'\xed:\x80\x00\x00\x01\x00\x01\x00\x00\x00\x01\x0f_acMe-cHaLlenge\tLOcaLHOSt\x07ExaMPle\x03com\x00\x00\x10\x00\x01\xc0\x0c\x00\x10\x00\x01\x00\x00\x00\x00\x00\x04\x03foo\x00\x00) \x00\x00\x00\x00\x00\x00\x00'
